@@ -16,13 +16,15 @@
 #include "common.h"
 
 #define NPKIT_GET_GPU_TIMESTAMP wall_clock64
+#define NPKIT_GET_CPU_TIMESTAMP_FROM_BLOCK \
+  __atomic_load_n(reinterpret_cast<uint64_t*>((uint8_t *)ncclShmem.comm.cpuTimestamp + 128*blockIdx.x), __ATOMIC_RELAXED)
 
 
 class NpKit {
  public:
-  static const uint64_t kNumGpuEventBuffers = 512;
+  static const uint64_t kNumGpuEventBuffers = 1024;
 
-  static const uint64_t kNumCpuEventBuffers = 32;
+  static const uint64_t kNumCpuEventBuffers = 64;
 
   static ncclResult_t Init(int rank);
 
