@@ -32,7 +32,7 @@ namespace RcclUnitTesting
 
     // Configuration
     std::vector<ncclFunc_t>     const funcTypes       = {ncclCollAllReduce};
-    std::vector<ncclDataType_t> const dataTypes       = {ncclFloat16, ncclFloat64};
+    std::vector<ncclDataType_t> const dataTypes       = {ncclFloat16, ncclFloat64, ncclFp8E4M3, ncclFp8E5M2};
     std::vector<ncclRedOp_t>    const redOps          = {ncclMin};
     std::vector<int>            const roots           = {0};
     std::vector<int>            const numElements     = {12888};
@@ -70,7 +70,7 @@ namespace RcclUnitTesting
 
     // Configuration
     std::vector<ncclFunc_t>     const funcTypes       = {ncclCollAllReduce};
-    std::vector<ncclDataType_t> const dataTypes       = {ncclInt32};
+    std::vector<ncclDataType_t> const dataTypes       = {ncclInt32, ncclFp8E4M3, ncclFp8E5M2};
     std::vector<ncclRedOp_t>    const redOps          = {ncclMax};
     std::vector<int>            const roots           = {0};
     std::vector<int>            const numElements     = {393216, 12888, 384};
@@ -102,6 +102,33 @@ namespace RcclUnitTesting
     testBed.Finalize();
   }
 
+  TEST(AllReduce, Channels)
+  {
+    TestBed testBed;
+    if(testBed.ev.maxGpus >= 8) {
+      if(testBed.ev.isGfx94) {
+        // Configuration
+        std::vector<ncclFunc_t>     const funcTypes       = {ncclCollAllReduce};
+        std::vector<ncclDataType_t> const dataTypes       = {ncclBfloat16};
+        std::vector<ncclRedOp_t>    const redOps          = {ncclSum};
+        std::vector<int>            const roots           = {0};
+        std::vector<int>            const numElements     = {64 * 1024 * 1024, 1024};
+        std::vector<bool>           const inPlaceList     = {false};
+        std::vector<bool>           const managedMemList  = {false};
+        std::vector<bool>           const useHipGraphList = {false, true};
+        std::vector<char *>         const channelList     = {"84", "112"};
+        bool                        const enableSweep     = false; 
+        for (auto channel : channelList) {
+          setenv("NCCL_MIN_NCHANNELS", channel, 1);
+          testBed.RunSimpleSweep(funcTypes, dataTypes, redOps, roots, numElements,
+                                inPlaceList, managedMemList, useHipGraphList, enableSweep);
+          testBed.Finalize();
+          unsetenv("NCCL_MIN_NCHANNELS");
+        }
+      }
+    }
+  }
+
   TEST(AllReduce, ManagedMemGraph)
   {
     TestBed testBed;
