@@ -16,7 +16,7 @@ namespace RcclUnitTesting
 
     // Configuration
     std::vector<ncclFunc_t>     const funcTypes       = {ncclCollAllToAll};
-    std::vector<ncclDataType_t> const dataTypes       = {ncclFloat32};
+    std::vector<ncclDataType_t> const dataTypes       = {ncclFloat16, ncclFloat32};
     std::vector<ncclRedOp_t>    const redOps          = {ncclSum};
     std::vector<int>            const roots           = {0};
     std::vector<int>            const numElements     = {1048576, 1024};
@@ -35,7 +35,7 @@ namespace RcclUnitTesting
 
     // Configuration
     std::vector<ncclFunc_t>     const funcTypes       = {ncclCollAllToAll};
-    std::vector<ncclDataType_t> const dataTypes       = {ncclFloat64, ncclBfloat16};
+    std::vector<ncclDataType_t> const dataTypes       = {ncclFloat64, ncclBfloat16, ncclFp8E4M3, ncclFp8E5M2};
     std::vector<ncclRedOp_t>    const redOps          = {ncclSum};
     std::vector<int>            const roots           = {0};
     std::vector<int>            const numElements     = {5685};
@@ -85,4 +85,31 @@ namespace RcclUnitTesting
                            inPlaceList, managedMemList, useHipGraphList);
     testBed.Finalize();
   }
-}
+
+    TEST(AllToAll, Channels)
+  {
+    TestBed testBed;
+    if(testBed.ev.maxGpus >= 8) {
+      if(testBed.ev.isGfx94) {
+        // Configuration
+        std::vector<ncclFunc_t>     const funcTypes       = {ncclCollAllToAll};
+        std::vector<ncclDataType_t> const dataTypes       = {ncclBfloat16};
+        std::vector<ncclRedOp_t>    const redOps          = {ncclSum};
+        std::vector<int>            const roots           = {0};
+        std::vector<int>            const numElements     = {64 * 1024 * 1024, 1024};
+        std::vector<bool>           const inPlaceList     = {false};
+        std::vector<bool>           const managedMemList  = {false};
+        std::vector<bool>           const useHipGraphList = {false, true};
+        std::vector<char *>         const channelList     = {"112"}; 
+        bool                        const enableSweep     = false;
+        for (auto channel : channelList) {
+          setenv("NCCL_MIN_NCHANNELS", channel, 1);
+          testBed.RunSimpleSweep(funcTypes, dataTypes, redOps, roots, numElements,
+                                inPlaceList, managedMemList, useHipGraphList, enableSweep);
+          testBed.Finalize();
+          unsetenv("NCCL_MIN_NCHANNELS");
+        }
+      }
+    }
+  }
+}
\ No newline at end of file
