@@ -5,7 +5,7 @@
  ************************************************************************/
 
 #include "PtrUnion.hpp"
-
+#include "api_trace.h"
 namespace RcclUnitTesting
 {
   size_t DataTypeToBytes(ncclDataType_t const dataType)
@@ -14,6 +14,8 @@ namespace RcclUnitTesting
     {
     case ncclInt8:   return 1;
     case ncclUint8:  return 1;
+    case ncclFp8E4M3:return 1;
+    case ncclFp8E5M2:return 1;
     case ncclInt32:  return 4;
     case ncclUint32: return 4;
     case ncclInt64:  return 8;
@@ -40,26 +42,38 @@ namespace RcclUnitTesting
     return TEST_SUCCESS;
   }
 
-  ErrCode PtrUnion::AllocateGpuMem(size_t const numBytes, bool const useManagedMem)
+  ErrCode PtrUnion::AllocateGpuMem(size_t const numBytes, bool const useManagedMem, bool const userRegistered)
   {
     if (numBytes)
     {
-      if (useManagedMem)
+      if (userRegistered)
       {
-        if (hipMallocManaged(&I1, numBytes) != hipSuccess)
+        if (ncclMemAlloc((void**)&I1, numBytes) != ncclSuccess)
         {
-          ERROR("Unable to allocate managed memory of GPU memory (%lu bytes)\n", numBytes);
+          ERROR("Unable to allocate user managed GPU memory (%lu bytes)\n", numBytes);
           return TEST_FAIL;
         }
       }
       else
       {
-        if (hipMalloc(&I1, numBytes) != hipSuccess)
+        if (useManagedMem)
         {
-          ERROR("Unable to allocate memory of GPU memory (%lu bytes)\n", numBytes);
-          return TEST_FAIL;
+          if (hipMallocManaged(&I1, numBytes) != hipSuccess)
+          {
+            ERROR("Unable to allocate managed memory of GPU memory (%lu bytes)\n", numBytes);
+            return TEST_FAIL;
+          }
+        }
+        else
+        {
+          if (hipMalloc(&I1, numBytes) != hipSuccess)
+          {
+            ERROR("Unable to allocate memory of GPU memory (%lu bytes)\n", numBytes);
+            return TEST_FAIL;
+          }
         }
       }
+
     }
     return TEST_SUCCESS;
   }
@@ -78,11 +92,14 @@ namespace RcclUnitTesting
     return TEST_SUCCESS;
   }
 
-  ErrCode PtrUnion::FreeGpuMem()
+  ErrCode PtrUnion::FreeGpuMem(bool const userRegistered)
   {
     if (this->ptr != nullptr)
     {
-      hipFree(this->ptr);
+      if (userRegistered)
+        ncclMemFree(this->ptr);
+      else
+        hipFree(this->ptr);
       this->ptr = nullptr;
     }
     return TEST_SUCCESS;
@@ -160,10 +177,12 @@ namespace RcclUnitTesting
     case ncclUint32:   U4[idx] = valueI; break;
     case ncclInt64:    I8[idx] = valueI; break;
     case ncclUint64:   U8[idx] = valueI; break;
+    case ncclFp8E4M3:  F1[idx] = rccl_float8(valueF); break;
     case ncclFloat16:  F2[idx] = __float2half(static_cast<float>(valueF)); break;
     case ncclFloat32:  F4[idx] = valueF; break;
     case ncclFloat64:  F8[idx] = valueF; break;
-    case ncclBfloat16: B2[idx] = rccl_bfloat16(static_cast<float>(valueF)); break;
+    case ncclFp8E5M2:  B1[idx] = rccl_bfloat8(valueF); break;
+    case ncclBfloat16: B2[idx] = hip_bfloat16(static_cast<float>(valueF)); break;
     default:
       ERROR("Unsupported datatype\n");
       return TEST_FAIL;
@@ -181,9 +200,11 @@ namespace RcclUnitTesting
     case ncclUint32:   valueI = U4[idx]; break;
     case ncclInt64:    valueI = I8[idx]; break;
     case ncclUint64:   valueI = U8[idx]; break;
+    case ncclFp8E4M3:  valueF = float(F1[idx]); break; 
     case ncclFloat16:  valueF = __half2float(F2[idx]); break;
     case ncclFloat32:  valueF = F4[idx]; break;
     case ncclFloat64:  valueF = F8[idx]; break;
+    case ncclFp8E5M2:  valueF = float(B1[idx]); break; 
     case ncclBfloat16: valueF = B2[idx]; break;
     default:
       ERROR("Unsupported datatype\n");
@@ -211,9 +232,11 @@ namespace RcclUnitTesting
       case ncclUint32:   U4[idx] *= scalarsPerRank.U4[rank]; break;
       case ncclInt64:    I8[idx] *= scalarsPerRank.I8[rank]; break;
       case ncclUint64:   U8[idx] *= scalarsPerRank.U8[rank]; break;
+      case ncclFp8E4M3:  F1[idx]  = rccl_float8(F1[idx] * scalarsPerRank.F1[rank]); break;
       case ncclFloat16:  F2[idx]  = __float2half(__half2float(F2[idx]) * __half2float(scalarsPerRank.F2[rank])); break;
       case ncclFloat32:  F4[idx] *= scalarsPerRank.F4[rank]; break;
       case ncclFloat64:  F8[idx] *= scalarsPerRank.F8[rank]; break;
+      case ncclFp8E5M2:  B1[idx]  = rccl_bfloat8(B1[idx] * scalarsPerRank.B1[rank]); break;
       case ncclBfloat16: B2[idx] *= scalarsPerRank.B2[rank]; break;
       default:
         ERROR("Unsupported datatype\n");
@@ -244,9 +267,11 @@ namespace RcclUnitTesting
       case ncclUint32:   U4[idx] = ReduceOp(op, U4[idx], inputCpu.U4[idx]); break;
       case ncclInt64:    I8[idx] = ReduceOp(op, I8[idx], inputCpu.I8[idx]); break;
       case ncclUint64:   U8[idx] = ReduceOp(op, U8[idx], inputCpu.U8[idx]); break;
+      case ncclFp8E4M3:  F1[idx] = rccl_float8(ReduceOp(op, float(F1[idx]), float(inputCpu.F1[idx]))); break;
       case ncclFloat16:  F2[idx] = __float2half(ReduceOp(op, __half2float(F2[idx]), __half2float(inputCpu.F2[idx]))); break;
       case ncclFloat32:  F4[idx] = ReduceOp(op, F4[idx], inputCpu.F4[idx]); break;
       case ncclFloat64:  F8[idx] = ReduceOp(op, F8[idx], inputCpu.F8[idx]); break;
+      case ncclFp8E5M2:  B1[idx] = rccl_bfloat8(ReduceOp(op, float(B1[idx]), float(inputCpu.B1[idx]))); break;
       case ncclBfloat16: B2[idx] = ReduceOp(op, B2[idx], inputCpu.B2[idx]); break;
       default:
         ERROR("Unsupported datatype\n");
@@ -271,10 +296,12 @@ namespace RcclUnitTesting
       case ncclUint32:   U4[idx] /= divisor; break;
       case ncclInt64:    I8[idx] /= divisor; break;
       case ncclUint64:   U8[idx] /= divisor; break;
-      case ncclFloat16:  F2[idx]  = __float2half(__half2float(F2[idx])/divisor); break;
+      case ncclFp8E4M3:  F1[idx] = (rccl_float8((float)(F1[idx]) / divisor)); break;
+      case ncclFloat16:  F2[idx] = __float2half(__half2float(F2[idx])/divisor); break;
       case ncclFloat32:  F4[idx] /= divisor; break;
       case ncclFloat64:  F8[idx] /= divisor; break;
-      case ncclBfloat16: B2[idx] = (rccl_bfloat16((float)(B2[idx]) / divisor)); break;
+      case ncclFp8E5M2:  B1[idx] = (rccl_bfloat8((float)(B1[idx]) / divisor)); break;
+      case ncclBfloat16: B2[idx] = (hip_bfloat16((float)(B2[idx]) / divisor)); break;
       default:
         ERROR("Unsupported datatype\n");
         return TEST_FAIL;
@@ -301,9 +328,11 @@ namespace RcclUnitTesting
       case ncclUint32:  isMatch = (U4[idx] == expected.U4[idx]); break;
       case ncclInt64:   isMatch = (I8[idx] == expected.I8[idx]); break;
       case ncclUint64:  isMatch = (U8[idx] == expected.U8[idx]); break;
+      case ncclFp8E4M3: isMatch = (fabs(float(F1[idx]) - float(expected.F1[idx])) < 9e-2); break;
       case ncclFloat16: isMatch = (fabs(__half2float(F2[idx]) - __half2float(expected.F2[idx])) < 9e-2); break;
       case ncclFloat32: isMatch = (fabs(F4[idx] - expected.F4[idx]) < 1e-5); break;
       case ncclFloat64: isMatch = (fabs(F8[idx] - expected.F8[idx]) < 1e-12); break;
+      case ncclFp8E5M2: isMatch = (fabs(float(B1[idx]) - float(expected.B1[idx])) < 9e-2); break;
       case ncclBfloat16: isMatch = (fabs((float)B2[idx] - (float)expected.B2[idx]) < 9e-2); break;
       default:
         ERROR("Unsupported datatype\n");
@@ -328,12 +357,16 @@ namespace RcclUnitTesting
         ERROR("Expected output: %ld.  Actual output: %ld at index %lu\n", expected.I8[idx], I8[idx], idx); break;
       case ncclUint64:
         ERROR("Expected output: %lu.  Actual output: %lu at index %lu\n", expected.U8[idx], U8[idx], idx); break;
+      case ncclFp8E4M3:
+        ERROR("Expected output: %f.  Actual output: %f at index %lu\n", (float)expected.F1[idx], (float)F1[idx], idx); break;
       case ncclFloat16:
         ERROR("Expected output: %f.  Actual output: %f at index %lu\n", __half2float(expected.F2[idx]), __half2float(F2[idx]), idx); break;
       case ncclFloat32:
         ERROR("Expected output: %f.  Actual output: %f at index %lu\n", expected.F4[idx], F4[idx], idx); break;
       case ncclFloat64:
         ERROR("Expected output: %lf.  Actual output: %lf at index %lu\n", expected.F8[idx], F8[idx], idx); break;
+      case ncclFp8E5M2:
+        ERROR("Expected output: %f.  Actual output: %f at index %lu\n", (float)expected.B1[idx], (float)B1[idx], idx); break;
       case ncclBfloat16:
         ERROR("Expected output: %f.  Actual output: %f at index %lu\n", (float)expected.B2[idx], (float)B2[idx], idx); break;
       default:
@@ -358,9 +391,11 @@ namespace RcclUnitTesting
       case ncclUint32:   ss << U4[i]; break;
       case ncclInt64:    ss << I8[i]; break;
       case ncclUint64:   ss << U8[i]; break;
+      case ncclFp8E4M3:  ss << (float)F1[i]; break;
       case ncclFloat16:  ss << __half2float(F2[i]); break;
       case ncclFloat32:  ss << F4[i]; break;
       case ncclFloat64:  ss << F8[i]; break;
+      case ncclFp8E5M2:  ss << (float)B1[i]; break;
       case ncclBfloat16: ss << (float)B2[i]; break;
       default: break;
       }
